"""A CLI for the ramses_rf library.

ramses_rf is used to parse/process Honeywell's RAMSES-II RF protocol as used for
Heating (CH/DHW) and HVAC (ventilation) control.
"""

try:
    from ramses_cli.client import main

except ModuleNotFoundError:
    import os
    import sys

    sys.path.append(f"{os.path.dirname(__file__)}/src")

    from ramses_cli.client import main

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""RAMSES RF - Test the use_regex feature."""

import asyncio
from datetime import datetime as dt

import pytest
import serial  # type: ignore[import-untyped]

from ramses_rf import Command, Gateway, Packet
from ramses_tx.protocol import PortProtocol
from ramses_tx.schemas import SZ_INBOUND, SZ_OUTBOUND, SZ_USE_REGEX
from ramses_tx.transport import _str
from tests_rf.virtual_rf import VirtualRf

# other constants
ASSERT_CYCLE_TIME = 0.0005  # max_cycles_per_assert = max_sleep / ASSERT_CYCLE_TIME
DEFAULT_MAX_SLEEP = 0.1


RULES_INBOUND = {
    "63:262143": "04:262143",
    "(W.*) 1FC9 (...) 21": "\\g<1> 1FC9 \\g<2> 00",
    "--:------ --:------ 12:215819": "01:215819 --:------ 01:215819",
    "000C 006 02(04|08)00FFFFFF": "000C 006 02\\g<1>0013FFFF",
}

RULES_OUTBOUND = {
    "04:262143": "63:262143",
    "(W.*) 1FC9 (...) 00": "\\g<1> 1FC9 \\g<2> 21",
    "01:215819 --:------ 01:215819": "--:------ --:------ 12:215819",
}

RULES_COMBINED = {SZ_INBOUND: RULES_INBOUND, SZ_OUTBOUND: RULES_OUTBOUND}

TESTS_OUTBOUND = {  # sent, received by other
    " I 003 01:215819 --:------ 01:215819 0009 003 0000FF": " I 003 --:------ --:------ 12:215819 0009 003 0000FF",
    " I --- 04:262143 --:------ 04:262143 30C9 003 000713": " I --- 63:262143 --:------ 63:262143 30C9 003 000713",
    " I --- 04:262143 --:------ 01:182924 2309 003 0205DC": " I --- 63:262143 --:------ 01:182924 2309 003 0205DC",
    # NOTE: the below doesn't work with QoS, as expects a response pkt (would be an I)
    " W --- 30:098165 32:206251 --:------ 1FC9 006 0031DA797F75": " W --- 30:098165 32:206251 --:------ 1FC9 006 2131DA797F75",
}

TESTS_INBOUND = {  # sent by other, received
    " I 003 --:------ --:------ 12:215819 0009 003 0000FF": " I 003 01:215819 --:------ 01:215819 0009 003 0000FF",
    " I --- 63:262143 --:------ 63:262143 30C9 003 000713": " I --- 04:262143 --:------ 04:262143 30C9 003 000713",
    " I --- 63:262143 --:------ 01:182924 2309 003 0205DC": " I --- 04:262143 --:------ 01:182924 2309 003 0205DC",
    # NOTE: the below won't work with QoS - why? - still the case with refactored QoS
    " W --- 30:098165 32:206251 --:------ 1FC9 006 2131DA797F75": " W --- 30:098165 32:206251 --:------ 1FC9 006 0031DA797F75",
    "RP --- 01:182924 30:068640 --:------ 000C 006 020400FFFFFF": "RP --- 01:182924 30:068640 --:------ 000C 006 02040013FFFF",
    "RP --- 01:182924 30:068640 --:------ 000C 006 020800FFFFFF": "RP --- 01:182924 30:068640 --:------ 000C 006 02080013FFFF",
}

GWY_CONFIG = {
    "config": {
        "disable_discovery": True,
        "disable_qos": False,  # QoS is required for this test
        "enforce_known_list": False,
    }
}


# ### FIXTURES #########################################################################


@pytest.fixture(autouse=True)
def patches_for_tests(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("ramses_tx.protocol._DBG_DISABLE_IMPERSONATION_ALERTS", True)
    monkeypatch.setattr("ramses_tx.transport.MIN_INTER_WRITE_GAP", 0)


async def assert_this_pkt(
    gwy: Gateway, expected: Command, max_sleep: float = DEFAULT_MAX_SLEEP
) -> None:
    """Check, at the gateway layer, that the current packet is as expected."""
    for _ in range(int(max_sleep / ASSERT_CYCLE_TIME)):
        await asyncio.sleep(ASSERT_CYCLE_TIME)
        if gwy._this_msg and gwy._this_msg._pkt._frame == expected._frame:
            break
    assert gwy._this_msg and gwy._this_msg._pkt._frame == expected._frame


# ### TESTS ############################################################################


@pytest.mark.xdist_group(name="virt_serial")
async def test_regex_inbound_() -> None:
    """Check the inbound filters work as expected (this test works without QoS)."""

    rf = VirtualRf(2)

    # NOTE: the absence of reciprocal outbound tests is intentional
    config = GWY_CONFIG
    config["config"].update({SZ_USE_REGEX: {SZ_INBOUND: RULES_INBOUND}})  # type: ignore[dict-item]

    gwy_0 = Gateway(rf.ports[0], **config)
    ser_1 = serial.Serial(rf.ports[1])

    try:
        await gwy_0.start()
        assert gwy_0._protocol._transport

        for cmd, pkt in TESTS_INBOUND.items():
            ser_1.write(bytes(cmd.encode("ascii")) + b"\r\n")

            await assert_this_pkt(gwy_0, Command(pkt))

    finally:
        await gwy_0.stop()
        await rf.stop()


@pytest.mark.xdist_group(name="virt_serial")
async def test_regex_with_qos() -> None:
    """Check the in/outbound regex filters work as expected with QoS."""

    rf = VirtualRf(2)

    config = GWY_CONFIG
    config["config"].update({SZ_USE_REGEX: RULES_COMBINED})  # type: ignore[dict-item]

    gwy_0 = Gateway(rf.ports[0], **config)
    ser_1 = serial.Serial(rf.ports[1])

    if not isinstance(gwy_0._protocol, PortProtocol) or not gwy_0._protocol._context:
        await rf.stop()
        pytest.skip("QoS protocol not enabled")
    else:
        assert gwy_0._protocol._disable_qos is False  # needed for successful tests

    try:
        await gwy_0.start()
        assert gwy_0._protocol._transport

        _ = ser_1.read(ser_1.in_waiting)  # ser_1.flush() doesn't work?

        for before, after in TESTS_OUTBOUND.items():
            cmd = Command(before)
            if cmd.rx_header:  # we wont be getting any replies
                continue

            pkt_src = await gwy_0.async_send_cmd(cmd)  # , timeout=DEFAULT_WAIT_TIMEOUT)
            assert str(pkt_src) == before

            pkt_dst = Packet.from_port(dt.now(), _str(ser_1.read(ser_1.in_waiting)))
            assert str(pkt_dst) == after

        for before, after in TESTS_INBOUND.items():
            ser_1.write(bytes(before.encode("ascii")) + b"\r\n")
            await assert_this_pkt(gwy_0, Command(after))

    finally:
        await gwy_0.stop()
        await rf.stop()



#!/usr/bin/env python3
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Provide the base class for commands (constructed/sent packets) and packets.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from . import exceptions as exc
from .address import ALL_DEV_ADDR, NON_DEV_ADDR, Address, pkt_addrs
from .const import COMMAND_REGEX, DEV_ROLE_MAP, DEV_TYPE_MAP
from .ramses import (
    CODE_IDX_ARE_COMPLEX,
    CODE_IDX_ARE_NONE,
    CODE_IDX_ARE_SIMPLE,
    CODE_IDX_DOMAIN,
    CODES_ONLY_FROM_CTL,
    CODES_SCHEMA,
    CODES_WITH_ARRAYS,
    RQ_NO_PAYLOAD,
)

# TODO: add _has_idx (as func return only one type, or raise)

from .const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    Code,
)
from .const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    F8,
    F9,
    FA,
    FC,
    FF,
)

if TYPE_CHECKING:
    from .const import VerbT


_LOGGER = logging.getLogger(__name__)


HeaderT = str
PayloadT = str
_PktIdxT = str


class Frame:
    """The Frame class - used as a base by the Command and Packet classes.

    `RQ --- 01:078710 10:067219 --:------ 3220 005 0000050000`
    """

    src: Address  # Address | Device
    dst: Address  # Address | Device
    _addrs: tuple[Address, Address, Address]

    def __init__(self, frame: str) -> None:
        """Create a frame from a string.

        Will raise InvalidPacketError if it is invalid.
        """

        self._frame: str = frame
        if not COMMAND_REGEX.match(self._frame):
            raise exc.PacketInvalid(f"Bad frame: invalid structure: >>>{frame}<<<")

        fields = frame.lstrip().split(" ")

        self.verb: VerbT = frame[:2]  # type: ignore[assignment]
        self.seqn: str = fields[1]  # . frame[3:6]
        self.code: Code = fields[5]  # type: ignore[assignment]
        self.len_: str = fields[6]  # . frame[42:45]  FIXME: len_, _len & len(payload)/2
        self.payload: PayloadT = fields[7]  # frame[46:].split(" ")[0]
        self._len: int = int(len(self.payload) / 2)

        try:
            self.src, self.dst, *self._addrs = pkt_addrs(  # type: ignore[assignment]
                " ".join(fields[i] for i in range(2, 5))  # frame[7:36]
            )
        except exc.PacketInvalid as err:  # will be: InvalidAddrSetError
            raise exc.PacketInvalid("Bad frame: invalid address set") from err

        if len(self.payload) != int(self.len_) * 2:
            raise exc.PacketInvalid(
                f"Bad frame: invalid payload: "
                f"len({self.payload}) is not int('{self.len_}' * 2))"
            )

        self._ctx_: bool | str = None  # type: ignore[assignment]
        self._hdr_: str = None  # type: ignore[assignment]
        self._idx_: bool | str = None  # type: ignore[assignment]

        self._has_array_: bool = None  # type: ignore[assignment]
        self._has_ctl_: bool = None  # type: ignore[assignment]  # TODO: remove
        self._has_payload_: bool = None  # type: ignore[assignment]

        self._repr: str = None  # type: ignore[assignment]

    # FIXME: this is messy
    def _validate(self, *, strict_checking: bool = False) -> None:
        """Validate the frame: it may be a cmd or a (response) pkt.

        Raise an exception InvalidPacketError (InvalidAddrSetError) if it is not valid.
        """

        if len(self._frame[46:].split(" ")[0]) != int(self._frame[42:45]) * 2:
            raise exc.PacketInvalid("Bad frame: Payload length mismatch")

        try:
            # self.src, self.dst, *self._addrs = pkt_addrs(self._frame[7:36])
            src, dst, *addrs = pkt_addrs(self._frame[7:36])
        except exc.PacketInvalid as err:  # will be: InvalidAddrSetError
            raise exc.PacketInvalid("Bad frame: Invalid address set") from err

        if not strict_checking:
            return

        try:  # Strict checking: helps users avoid to constructing bad commands
            if addrs[0] == NON_DEV_ADDR:
                assert self.verb == I_, "wrong verb or dst addr should be present"
            elif addrs[2] == NON_DEV_ADDR:
                assert (
                    self.verb == I_ or src is not dst
                ), "wrong verb or dst addr should not be src"
            elif addrs[0] is addrs[2]:
                assert self.verb == I_, "wrong verb or dst addr should not be src"
            else:
                assert self.verb in (I_, W_), "wrong verb or dst addr should be src"
        except AssertionError as err:
            raise exc.PacketInvalid(f"Bad frame: Invalid address set: {err}") from err

    def __repr__(self) -> str:
        """Return a unambiguous string representation of this object."""

        if self._repr is None:
            self._repr = " ".join(  # type: ignore[unreachable]
                (
                    self.verb,
                    self.seqn,
                    *(repr(a) for a in self._addrs),
                    self.code,
                    self.len_,
                    self.payload,
                )
            )
        return self._repr

    def __str__(self) -> str:
        """Return a brief readable string representation of this object."""
        # repr(self) == repr(cls(str(self)))

        try:
            return f"{self!r} # {self._hdr}"  # code|ver|device_id|context
        except AttributeError as err:
            return f"{self!r} < {err}"

    def __eq__(self, other: object) -> bool:
        if not hasattr(other, "_frame"):
            return NotImplemented
        return self._frame[4:] == other._frame[4:]  # type: ignore[no-any-return]

    @property
    def _has_array(self) -> None | bool:  # TODO: a mess - has false negatives
        """Return the True if the payload is an array, False otherwise.

        May return false negatives (e.g. arrays of length 1), and None if undetermined.

        An example of a false negative is evohome with only one zone (i.e. the periodic
        2309/30C9/000A packets).
        """

        if self._has_array_ is not None:  # HACK: overriden by detect_array(msg, prev)
            return self._has_array_

        # False -ves (array length is 1) are an acceptable compromise to extensive checking

        # .W --- 01:145038 34:092243 --:------ 1FC9 006 07230906368E
        # .I --- 01:145038 --:------ 01:145038 1FC9 018 07000806368E-FC3B0006368E-071FC906368E
        # .I --- 01:145038 --:------ 01:145038 1FC9 018 FA000806368E-FC3B0006368E-FA1FC906368E
        # .I --- 34:092243 --:------ 34:092243 1FC9 030 0030C9896853-002309896853-001060896853-0010E0896853-001FC9896853
        if self.code == Code._1FC9:  # type: ignore[unreachable]
            self._has_array_ = self.verb != RQ  # safe to treat all as array, even len=1
            return self._has_array_  # don't do any checks for 1FC9 (they will fail)

        elif self.verb != I_ or self.code not in CODES_WITH_ARRAYS:
            self._has_array_ = False

        elif self._len != CODES_WITH_ARRAYS[self.code][0]:  # NOTE: can be false -ves
            a, b = divmod(self._len, CODES_WITH_ARRAYS[self.code][0])
            self._has_array_ = a > 0 and b == 0

        elif (
            self.code in (Code._22C9, Code._3150)
            and self.src.type == DEV_TYPE_MAP.UFC
            and self.dst is self.src
            and self.payload[:1] != "F"
        ):
            self._has_array_ = True

        # elif self.code == Code._000C:  # anachronism: variable array length
        #     return

        else:
            self._has_array_ = False

        if self._has_array_:
            len_ = CODES_WITH_ARRAYS[self.code][0]

            assert (
                self._len % len_ == 0
            ), f"{self} < array has length ({self._len}) that is not multiple of {len_}"
            assert (
                self.src.type in (DEV_TYPE_MAP.DTS, DEV_TYPE_MAP.DT2)
                or self.src == self.dst  # DEX
            ), f"{self} < array is from a non-controller (01)"
            assert (
                self.src.type not in (DEV_TYPE_MAP.DTS, DEV_TYPE_MAP.DT2)
                or self.dst.id == NON_DEV_ADDR.id  # DEX
            ), f"{self} < array is from a non-controller (02)"

        # .I --- 10:040239 01:223036 --:------ 0009 003 000000        # not array
        # .I --- 01:102458 --:------ 01:102458 0009 006 FC01FF-F901FF
        # .I --- 01:145038 --:------ 01:145038 0009 006 FC00FF-F900FF
        # .I 034 --:------ --:------ 12:126457 2309 006 017EFF-027EFF
        # .I --- 01:223036 --:------ 01:223036 000A 012 081001F40DAC-091001F40DAC  # 2nd fragment
        # .I 024 --:------ --:------ 12:126457 000A 012 010001F40BB8-020001F40BB8
        # .I --- 02:044328 --:------ 02:044328 22C9 018 0001F40A2801-0101F40A2801-0201F40A2801
        # .I --- 23:100224 --:------ 23:100224 2249 007 007EFF7EFFFFFF  # can have 2 zones
        # .I --- 02:044328 --:------ 02:044328 22C9 018 0001F40A2801-0101F40A2801-0201F40A2801
        # .I --- 02:001107 --:------ 02:001107 3150 010 007A-017A-027A-036A-046A

        return self._has_array_

    @property
    def _has_ctl(self) -> None | bool:
        """Return True if the packet is to/from a controller."""

        # NB: the difference between these (_has_ctl, src, dst, -:-) and above (_has_ctl, src, -:-, src)
        # 2000-01-01T03:00:00.000000 ...  I --- 37:123456 --:------ 37:123456 31DA 030 00C8400518646427102AF82EE031FFFFFFC800C8C83FFF64640AFF0AFF00
        # 2022-11-20T08:32:06.904058 063 RP --- 32:134446 37:171685 --:------ 31DA 030 00EF007FFF2F1B0226069A07EEFFE4F8000038988F0000EFEF1F2420FC00
        # Maybe only use this for CH/DHW, and not HVAC?

        if self._has_ctl_ is not None:
            return self._has_ctl_

        # TODO: handle RQ/RP to/from HGI/RFG, handle HVAC

        if {self.src.type, self.dst.type} & {  # type: ignore[unreachable]
            DEV_TYPE_MAP.CTL,
            DEV_TYPE_MAP.UFC,
            DEV_TYPE_MAP.PRG,
        }:  # DEX
            _LOGGER.debug(f"{self} # HAS controller (10)")
            self._has_ctl_ = True

        # .I --- 12:010740 --:------ 12:010740 30C9 003 0008D9 # not ctl
        elif self.dst is self.src:  # (not needed?) & self.code == I_:
            _LOGGER.debug(
                f"{self} < "
                + (
                    "HAS"
                    if self.code in CODES_ONLY_FROM_CTL + (Code._31D9, Code._31DA)
                    else "no"
                )
                + " controller (20)"
            )
            self._has_ctl_ = any(
                (
                    self.code == Code._3B00 and self.payload[:2] == FC,
                    self.code in CODES_ONLY_FROM_CTL + (Code._31D9, Code._31DA),
                )
            )

        # .I --- --:------ --:------ 10:050360 1FD4 003 002ABE # no ctl
        # .I 095 --:------ --:------ 12:126457 1F09 003 000BC2 # HAS ctl
        # .I --- --:------ --:------ 20:001473 31D9 003 000001 # ctl? (HVAC)
        elif self.dst.id == NON_DEV_ADDR.id:
            _LOGGER.debug(f"{self} # HAS controller (21)")
            self._has_ctl_ = self.src.type != DEV_TYPE_MAP.OTB  # DEX

        # .I --- 10:037879 --:------ 12:228610 3150 002 0000   # HAS ctl
        # .I --- 04:029390 --:------ 12:126457 1060 003 01FF01 # HAS ctl
        elif self.dst.type in (DEV_TYPE_MAP.DTS, DEV_TYPE_MAP.DT2):  # DEX
            _LOGGER.debug(f"{self} # HAS controller (22)")
            self._has_ctl_ = True

        # RQ --- 30:258720 10:050360 --:------ 3EF0 001 00           # UNKNOWN (99)
        # RP --- 10:050360 30:258720 --:------ 3EF0 006 000011010A1C # UNKNOWN (99)

        # RQ --- 18:006402 13:049798 --:------ 1FC9 001 00
        # RP --- 13:049798 18:006402 --:------ 1FC9 006 003EF034C286
        # RQ --- 30:258720 10:050360 --:------ 22D9 001 00
        # RP --- 10:050360 30:258720 --:------ 22D9 003 0003E8
        # RQ --- 30:258720 10:050360 --:------ 3220 005 0000120000
        # RP --- 10:050360 30:258720 --:------ 3220 005 0040120166
        # RQ --- 30:258720 10:050360 --:------ 3EF0 001 00
        # RP --- 10:050360 30:258720 --:------ 3EF0 006 000011010A1C

        # .I --- 34:021943 63:262142 --:------ 10E0 038 000001C8380A01... # unknown
        # .I --- 32:168090 30:082155 --:------ 31E0 004 0000C800          # unknown

        if self._has_ctl_ is None:
            # if DEV_MODE and DEV_TYPE_MAP.HGI not in (
            #     self.src.type,
            #     self.dst.type,
            # ):  # DEX
            #     _LOGGER.warning(f"{self} # has_ctl - undetermined (99)")
            self._has_ctl_ = False

        return self._has_ctl_

    @property
    def _has_idx(self) -> bool:
        """Return True if the payload has an index (or has an array), False otherwise."""

        return self._idx is not False

    @property
    def _has_payload(self) -> bool:
        """Return True if the packet has a non-null payload, False otherwise.

        May return false positives. The payload may still have an idx.
        """

        if self._has_payload_ is not None:
            return self._has_payload_

        self._has_payload_ = not any(  # type: ignore[unreachable]
            (
                self._len == 1,
                self.verb == RQ and self.code in RQ_NO_PAYLOAD,
                self.verb == RQ and self._len == 2 and self.code != Code._0016,
                # self.verb == RQ and self._len == 2 and self.code in (
                #   Code._2309, Code._2349, Code._3EF1
                # ),
            )
        )

        return self._has_payload_

    def _force_has_array(self) -> None:
        self._has_array_ = True
        self._ctx_ = None  # type: ignore[assignment]
        self._hdr_ = None  # type: ignore[assignment]
        self._idx_ = None  # type: ignore[assignment]

    @property
    def _is_fragment(self) -> bool:
        """Return True is the payload *could* be a fragment, False otherwise."""
        # .I 036 --:------ --:------ 12:126457 000A 012 010001F40BB8-020001F40BB8  # max 2 zones

        return (
            self.code in (Code._000A, Code._22C9)
            and self.verb == I_
            and self.src is self.dst
            # and self._has_array  # not needed
        ) or (self.code == Code._0404 and self.verb in (I_, RP))

    @property
    def _ctx(self) -> bool | str:  # incl. self._idx
        """Return the payload's full context, if any (e.g. for 0404: zone_idx/frag_idx).

        Used to store packets in the entity's message DB. It is a superset of _idx.
        """

        if self._ctx_ is not None:
            return self._ctx_

        if self.code in (  # type: ignore[unreachable]
            Code._0005,
            Code._000C,
        ):  # zone_idx, zone_type (device_role)
            self._ctx_ = self.payload[:4]
        elif self.code == Code._0404:  # zone_idx, frag_idx
            self._ctx_ = self._idx + self.payload[10:12]
        else:
            self._ctx_ = self._idx
        return self._ctx_

    @property
    def _hdr(self) -> HeaderT:  # incl. self._ctx
        """Return the QoS header (fingerprint) of this packet (i.e. device_id/code/hdr).

        Used for QoS (timeouts, retries), callbacks, etc.
        """

        if self._hdr_ is not None:
            return self._hdr_

        # FIXME: HACK: sometimes RecursionError
        self._hdr_ = "|".join((self.code, self.verb))  # type: ignore[unreachable]
        self._hdr_ = pkt_header(self)
        return self._hdr_

    @property
    def _idx(self) -> bool | str:  # FIXME: a mess
        """Return the payload's index, if any (e.g. zone_idx, domain_id  or log_idx).

        Used to route a packet to the correct entity's (i.e. zone/domain) msg handler.
        """

        if self._idx_ is not None:
            return self._idx_

        self._idx_ = _pkt_idx(self) or False  # type: ignore[unreachable]
        return self._idx_


# TODO: a mess - has false negatives
def _pkt_idx(pkt: Frame) -> None | bool | str:  # _has_array, _has_ctl
    """Return the payload's 2-byte context (e.g. zone_idx, domain_id or log_idx).

    May return a 2-byte string (usu. pkt.payload[:2]), or:
    - False if there is no context at all
    - True if the payload is an array
    - None if it is indeterminable
    """
    # The three iterables (none, simple, complex) are mutex

    # FIXME: 0016 is broken

    # mutex 2/4, CODE_IDX_COMPLEX: are not payload[:2]
    if pkt.code == Code._0005:
        return pkt._has_array

    # .I --- 10:040239 01:223036 --:------ 0009 003 000000
    if pkt.code == Code._0009 and pkt.src.type == DEV_TYPE_MAP.OTB:  # DEX
        return False

    if pkt.code == Code._000C:  # zone_idx/domain_id (complex, payload[0:4])
        if pkt.payload[2:4] == DEV_ROLE_MAP.APP:  # "000F"
            return str(FC)  # mypy
        if pkt.payload[0:4] == f"01{DEV_ROLE_MAP.HTG}":  # "010E"
            return str(F9)  # mypy
        if pkt.payload[2:4] in (
            DEV_ROLE_MAP.DHW,
            DEV_ROLE_MAP.HTG,
        ):  # "000D", "000E"
            return str(FA)  # mypy
        return pkt.payload[:2]

    if pkt.code == Code._0404:  # assumes only 1 DHW zone (can be 2, but never seen)
        return "HW" if pkt.payload[2:4] == "23" else pkt.payload[:2]

    if pkt.code == Code._0418:  # log_idx (payload[4:6])
        return pkt.payload[4:6]

    if pkt.code == Code._1100:  # TODO; can do in parser
        return pkt.payload[:2] if pkt.payload[:1] == "F" else False  # only FC

    if pkt.code == Code._3220:  # msg_id/data_id (payload[4:6])
        return pkt.payload[4:6]

    if pkt.code in CODE_IDX_ARE_COMPLEX:  # these should be handled above
        raise NotImplementedError(f"{pkt} # CODE_IDX_COMPLEX")  # a coding error

    # mutex 1/4, CODE_IDX_NONE: always returns False
    if pkt.code in CODE_IDX_ARE_NONE:  # returns False
        if (
            CODES_SCHEMA[pkt.code].get(pkt.verb, "")[:3] == "^00"
            and pkt.payload[:2] != "00"
        ):
            raise exc.PacketPayloadInvalid(
                f"Packet idx is {pkt.payload[:2]}, but expecting no idx (00) (0xAA)"
            )
        return False

    # mutex 3/4, CODE_IDX_SIMPLE: potentially some false -ves?
    if pkt._has_array:
        return True  # excludes len==1 for 000A, 2309, 30C9

    # TODO: is this needed?: exceptions to CODE_IDX_SIMPLE
    if pkt.payload[:2] in (F8, F9, FA, FC):  # TODO: F6, F7?, FB, FD
        if pkt.code not in CODE_IDX_DOMAIN:
            raise exc.PacketPayloadInvalid(
                f"Packet idx is {pkt.payload[:2]}, but not expecting a domain id"
            )
        return pkt.payload[:2]

    if (
        pkt._has_ctl  # TODO: exclude HVAC?
    ):  # risk of false -ves, TODO: pkt.src.type == DEV_TYPE_MAP.HGI too?  # DEX
        # 02:    22C9: would be picked up as an array, if len==1 counted
        # 03:    # .I 028 03:094242 --:------ 03:094242 30C9 003 010B22  # ctl
        # 12/22: 000A|1030|2309|30C9 from (addr0 --:), 1060|3150 (addr0 04:)
        # 23:    0009|10A0
        return pkt.payload[:2]  # tcs._max_zones checked elsewhere

    if pkt.code in (Code._31D9, Code._31DA):
        return pkt.payload[:2]

    if pkt.payload[:2] != "00":
        raise exc.PacketPayloadInvalid(
            f"Packet idx is {pkt.payload[:2]}, but expecting no idx (00) (0xAB)"
        )  # TODO: add a test for this

    if pkt.code in CODE_IDX_ARE_SIMPLE:
        return None  # False  # TODO: return None (less precise) or risk false -ves?

    # mutex 4/4, CODE_IDX_UNKNOWN: an unknown code
    _LOGGER.info(f"{pkt} # Unable to determine payload index (is probably OK)")
    return None


def pkt_header(pkt: Frame, /, rx_header: bool = False) -> None | HeaderT:
    """Return the header of a packet (all packets have a header).

    Used for QoS, and others.

    For rx_header=True, return instead the header of the response packet, if one is
    expected, otherwise return None.

    Examples include:
     I --- 04:155407 --:------ 04:155407 30C9 003 00092F                   # 30C9| I|04:155407
    RP --- 01:223036 18:005567 --:------ 2349 007 0404B000FFFFFF           # 2349|RP|01:223036|04
     I --- 01:223036 --:------ 01:223036 3B00 002 FCC8                     # 3B00| I|01:223036|FC
    RP --- 01:223036 18:005567 --:------ 000C 012 020800125F91020800125F8D # 000C|RP|01:223036|0208
     I --- 01:223036 --:------ 01:223036 2309 030 00-0640 01-03E8 02-03... # 2309| I|01:223036 (True)
    """

    if pkt.code == Code._1FC9:
        # .I --- 34:021943 --:------ 34:021943 1FC9 024 00-2309-8855B7 00-1FC9-8855B7
        # .W --- 01:145038 34:021943 --:------ 1FC9 006 00-2309-06368E  # wont know src until it arrives
        # .I --- 34:021943 01:145038 --:------ 1FC9 006 00-2309-8855B7
        if not rx_header:
            device_id = ALL_DEV_ADDR.id if pkt.src == pkt.dst else pkt.dst.id
            return "|".join((pkt.code, pkt.verb, device_id))
        if pkt.src == pkt.dst:  # and pkt.verb == I_:
            return "|".join((pkt.code, W_, pkt.src.id))
        if pkt.verb == W_:  # and pkt.src != pkt.dst:
            return "|".join((pkt.code, I_, pkt.src.id))  # TODO: why not pkt.dst?
        # if pkt.verb == RQ:  # and pkt.src != pkt.dst:  # TODO: this breaks things
        #     return "|".join((pkt.code, RP, pkt.dst.id))
        return None

    # RQ and W use the dst.id rather than the src.id, as:
    # - cmd.src could be 18:000730, and echo .src will have changed to (say) 18:123456
    # - cmd.dst is the effector

    if rx_header:
        if pkt.verb in (I_, RP) or pkt.src == pkt.dst:  # say: xxxx| W|00:000000|xx
            return None  # no response expected
        header = "|".join((pkt.code, RP if pkt.verb == RQ else I_, pkt.dst.id))

    elif pkt.verb in (I_, RP) or pkt.src == pkt.dst:
        header = "|".join((pkt.code, pkt.verb, pkt.src.id))

    else:
        header = "|".join((pkt.code, pkt.verb, pkt.dst.id))

    try:
        return f"{header}|{pkt._ctx}" if isinstance(pkt._ctx, str) else header
    except AssertionError:
        return header



#!/usr/bin/env python3
"""RAMSES RF - RAMSES-II compatible packet protocol finite state machine."""

from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable, Coroutine
from datetime import datetime as dt
from queue import Empty, Full, PriorityQueue
from threading import Lock
from typing import TYPE_CHECKING, Any, Final, TypeAlias

from . import exceptions as exc
from .address import HGI_DEVICE_ID
from .command import Command
from .const import (
    DEFAULT_BUFFER_SIZE,
    DEFAULT_ECHO_TIMEOUT,
    DEFAULT_RPLY_TIMEOUT,
    MAX_RETRY_LIMIT,
    MAX_SEND_TIMEOUT,
    Code,
    Priority,
)
from .packet import Packet
from .typing import QosParams

if TYPE_CHECKING:
    from .protocol import RamsesProtocolT
    from .transport import RamsesTransportT
    from .typing import ExceptionT

#
# NOTE: All debug flags should be False for deployment to end-users
_DBG_MAINTAIN_STATE_CHAIN: Final[bool] = False  # maintain Context._prev_state
_DBG_USE_STRICT_TRANSITIONS: Final[bool] = False

_LOGGER = logging.getLogger(__name__)


#######################################################################################

_FutureT: TypeAlias = asyncio.Future[Packet]
_QueueEntryT: TypeAlias = tuple[Priority, dt, Command, QosParams, _FutureT]


class ProtocolContext:
    SEND_TIMEOUT_LIMIT = MAX_SEND_TIMEOUT

    def __init__(
        self,
        protocol: RamsesProtocolT,
        /,
        *,
        echo_timeout: float = DEFAULT_ECHO_TIMEOUT,
        reply_timeout: float = DEFAULT_RPLY_TIMEOUT,
        max_retry_limit: int = MAX_RETRY_LIMIT,
        max_buffer_size: int = DEFAULT_BUFFER_SIZE,
    ) -> None:
        self._protocol = protocol
        self.echo_timeout = echo_timeout
        self.reply_timeout = reply_timeout
        self.max_retry_limit = min(max_retry_limit, MAX_RETRY_LIMIT)
        self.max_buffer_size = min(max_buffer_size, DEFAULT_BUFFER_SIZE)

        self._loop = protocol._loop
        self._lock = Lock()  # FIXME: threading lock, or asyncio lock?
        self._fut: _FutureT | None = None
        self._que: PriorityQueue[_QueueEntryT] = PriorityQueue(
            maxsize=self.max_buffer_size
        )

        self._expiry_timer: asyncio.Task[None] | None = None
        self._multiplier = 0
        self._state: _ProtocolStateT = None  # type: ignore[assignment]

        # TODO: pass this over as an instance paramater
        self._send_fnc: Callable[[Command], Coroutine[Any, Any, None]] = None  # type: ignore[assignment]

        self._cmd: Command | None = None
        self._qos: QosParams | None = None
        self._cmd_tx_count: int = 0  # was: None
        self._cmd_tx_limit: int = 0

        self.set_state(Inactive)

    def __repr__(self) -> str:
        msg = f"<ProtocolContext state={repr(self._state)[21:-1]}"
        if self._cmd is None:
            return msg + ">"
        if self._cmd_tx_count == 0:  # was: is None
            return msg + ", tx_count=0/0>"
        return msg + f", tx_count={self._cmd_tx_count}/{self._cmd_tx_limit}>"

    @property
    def state(self) -> _ProtocolStateT:
        return self._state

    @property
    def is_sending(self) -> bool:  # TODO: remove asserts
        if isinstance(self._state, WantEcho | WantRply):
            assert self._cmd is not None, f"{self}: Coding error"  # mypy hint
            assert self._qos is not None, f"{self}: Coding error"  # mypy hint
            assert self._fut is not None, f"{self}: Coding error"  # mypy hint
            return True

        assert self._cmd is None, f"{self}: Coding error"  # mypy hint
        assert self._qos is None, f"{self}: Coding error"  # mypy hint
        assert (
            self._fut is None or self._fut.done()
        ), f"{self}: Coding error"  # mypy hint
        return False

    def set_state(
        self,
        state_class: _ProtocolStateClassT,
        expired: bool = False,
        timed_out: bool = False,
        exception: Exception | None = None,
        result: Packet | None = None,
    ) -> None:
        async def expire_state_on_timeout() -> None:
            # a separate coro, so can be spawned off with create_task()

            assert self._cmd is not None  # mypy

            assert isinstance(
                self.is_sending, bool
            ), f"{self}: Coding error"  # TODO: remove
            assert self._cmd_tx_count > 0, f"{self}: Coding error"  # TODO: remove

            if isinstance(self._state, WantEcho):  # otherwise is WantRply
                delay = self.echo_timeout * (2**self._multiplier)
            # elif self._cmd.code == Code._0404:
            #     delay = self.reply_timeout * (2**self._multiplier) * 2
            else:  # isinstance(self._state, WantRply):
                delay = self.reply_timeout * (2**self._multiplier)

            # assuming success, multiplier can be decremented...
            self._multiplier, old_val = max(0, self._multiplier - 1), self._multiplier

            await asyncio.sleep(delay)  # ideally, will be interrupted by wait_for()

            # nope, was not successful, so multiplier should be incremented...
            self._multiplier = min(3, old_val + 1)

            if isinstance(self._state, WantEcho):
                _LOGGER.warning("TOUT.. = %s: echo_timeout=%s", self, delay)
            else:  # isinstance(self._state, WantRply):
                _LOGGER.warning("TOUT.. = %s: rply_timeout=%s", self, delay)

            assert isinstance(
                self.is_sending, bool
            ), f"{self}: Coding error"  # TODO: remove

            # Timer has expired, can we retry or are we done?
            assert isinstance(self._cmd_tx_count, int)

            if self._cmd_tx_count < self._cmd_tx_limit:
                self.set_state(WantEcho, timed_out=True)
            else:
                self.set_state(IsInIdle, expired=True)

            assert isinstance(
                self.is_sending, bool
            ), f"{self}: Coding error"  # TODO: remove

        def effect_state(timed_out: bool) -> None:
            """Take any actions indicated by state, and optionally set expiry timer."""
            # a separate function, so can be spawned off with call_soon()

            assert isinstance(
                self.is_sending, bool
            ), f"{self}: Coding error"  # TODO: remove

            if timed_out:
                assert self._cmd is not None, f"{self}: Coding error"  # mypy hint
                self._send_cmd(self._cmd, is_retry=True)

            if isinstance(self._state, IsInIdle):
                self._loop.call_soon_threadsafe(self._check_buffer_for_cmd)

            elif isinstance(self._state, WantRply) and not self._qos.wait_for_reply:  # type: ignore[union-attr]
                self.set_state(IsInIdle, result=self._state._echo_pkt)

            elif isinstance(self._state, WantEcho | WantRply):
                self._expiry_timer = self._loop.create_task(expire_state_on_timeout())

        if self._expiry_timer is not None:
            self._expiry_timer.cancel()
            self._expiry_timer = None

        # when _fut.done(), three possibilities:
        #  _fut.set_result()
        #  _fut.set_exception()
        #  _fut.cancel() (via a wait_for())

        # Changing the order of the following is fraught with danger
        if self._fut is None:  # logging only - IsInIdle, Inactive
            _LOGGER.debug("BEFORE = %s", self)
            assert self._cmd is None, f"{self}: Coding error"  # mypy hint
            assert isinstance(
                self._state, IsInIdle | Inactive | None
            ), f"{self}: Coding error"  # mypy hint

        elif self._fut.cancelled():  # by send_cmd(qos.timeout)
            _LOGGER.debug("BEFORE = %s: expired=%s (global)", self, expired)
            assert self._cmd is not None, f"{self}: Coding error"  # mypy hint
            assert isinstance(
                self._state, WantEcho | WantRply
            ), f"{self}: Coding error"  # mypy hint

        elif exception:
            _LOGGER.debug("BEFORE = %s: exception=%s", self, exception)
            assert (
                not self._fut.done()
            ), f"{self}: Coding error ({self._fut})"  # mypy hint
            assert isinstance(
                self._state, WantEcho | WantRply
            ), f"{self}: Coding error"  # mypy hint
            self._fut.set_exception(exception)  # apologise to the sender

        elif result:
            _LOGGER.debug("BEFORE = %s: result=%s", self, result._hdr)
            assert (
                not self._fut.done()
            ), f"{self}: Coding error ({self._fut})"  # mypy hint
            assert isinstance(
                self._state, WantEcho | WantRply
            ), f"{self}: Coding error"  # mypy hint
            self._fut.set_result(result)

        elif expired:  # by expire_state_on_timeout(echo_timeout/reply_timeout)
            _LOGGER.debug("BEFORE = %s: expired=%s", self, expired)
            assert (
                not self._fut.done()
            ), f"{self}: Coding error ({self._fut})"  # mypy hint
            assert isinstance(
                self._state, WantEcho | WantRply
            ), f"{self}: Coding error"  # mypy hint
            self._fut.set_exception(
                exc.ProtocolSendFailed(f"{self}: Exceeded maximum retries")
            )

        else:  # logging only - WantEcho, WantRply
            _LOGGER.debug("BEFORE = %s", self)
            assert (
                self._fut is None or self._fut.cancelled() or not self._fut.done()
            ), f"{self}: Coding error ({self._fut})"  # mypy hint
            # sert isinstance(self._state, WantEcho | WantRply), f"{self}: Coding error"  # mypy hint

        prev_state = self._state  # for _DBG_MAINTAIN_STATE_CHAIN

        self._state = state_class(self)  # keep atomic with tx_count / tx_limit calcs

        if _DBG_MAINTAIN_STATE_CHAIN:  # for debugging
            # tattr(prev_state, "_next_state", self._state)
            setattr(self._state, "_prev_state", prev_state)  # noqa: B010

        if timed_out:  # isinstance(self._state, WantEcho):
            assert isinstance(
                self._cmd_tx_count, int
            ), f"{self}: Coding error"  # mypy hint
            self._cmd_tx_count += 1

        elif isinstance(self._state, WantEcho):
            assert self._qos is not None, f"{self}: Coding error"  # mypy hint
            # self._cmd_tx_limit = min(self._qos.max_retries, self.max_retry_limit) + 1
            self._cmd_tx_count = 1

        elif not isinstance(self._state, WantRply):  # IsInIdle, IsInactive
            self._cmd = self._qos = None
            self._cmd_tx_count = 0  # was: = None

        assert isinstance(self.is_sending, bool)  # TODO: remove

        # remaining code spawned off with a call_soon(), so early return to caller
        self._loop.call_soon_threadsafe(effect_state, timed_out)  # calls expire_state

        if not isinstance(self._state, WantRply):
            _LOGGER.debug("AFTER. = %s", self)
            return

        assert self._qos is not None, f"{self}: Coding error"  # mypy hint
        _LOGGER.debug("AFTER. = %s: wait_for_reply=%s", self, self._qos.wait_for_reply)

    def connection_made(self, transport: RamsesTransportT) -> None:
        # may want to set some instance variables, according to type of transport
        self._state.connection_made()

    # TODO: Should we clear the buffer if connection is lost (and apoligise to senders?
    def connection_lost(self, err: ExceptionT | None) -> None:
        self._state.connection_lost()

    def pkt_received(self, pkt: Packet) -> Any:
        self._state.pkt_rcvd(pkt)

    def pause_writing(self) -> None:
        self._state.writing_paused()

    def resume_writing(self) -> None:
        self._state.writing_resumed()

    async def send_cmd(
        self,
        send_fnc: Callable[[Command], Coroutine[Any, Any, None]],  # TODO: remove
        cmd: Command,
        priority: Priority,
        qos: QosParams,
    ) -> Packet:
        self._send_fnc = send_fnc  # TODO: REMOVE: make per Context, not per Command

        if isinstance(self._state, Inactive):
            raise exc.ProtocolSendFailed(f"{self}: Send failed (no active transport?)")

        assert self._loop is asyncio.get_running_loop()  # BUG is here

        fut: _FutureT = self._loop.create_future()
        try:
            self._que.put_nowait((priority, dt.now(), cmd, qos, fut))
        except Full as err:
            fut.cancel()
            raise exc.ProtocolSendFailed(f"{self}: Send buffer overflow") from err

        if isinstance(self._state, IsInIdle):
            self._loop.call_soon_threadsafe(self._check_buffer_for_cmd)

        timeout = min(  # needs to be greater than worse-case via set_state engine
            qos.timeout, self.SEND_TIMEOUT_LIMIT
        )  # incl. time queued in buffer
        try:
            await asyncio.wait_for(fut, timeout=timeout)
        except TimeoutError as err:  # incl. fut.cancel()
            msg = f"{self}: Expired global timer after {timeout} sec"
            _LOGGER.warning(
                "TOUT.. = %s: send_timeout=%s (%s)", self, timeout, self._cmd is cmd
            )
            if self._cmd is cmd:  # NOTE: # this cmd may not yet be self._cmd
                self.set_state(
                    IsInIdle, expired=True
                )  # set_exception() will cause InvalidStateError
            raise exc.ProtocolSendFailed(msg) from err  # make msg *before* state reset

        try:
            return fut.result()
        except exc.ProtocolSendFailed:
            raise
        except (exc.ProtocolError, exc.TransportError) as err:  # incl. ProtocolFsmError
            raise exc.ProtocolSendFailed(f"{self}: Send failed: {err}") from err

    def _check_buffer_for_cmd(self) -> None:
        self._lock.acquire()
        assert isinstance(self.is_sending, bool), f"{self}: Coding error"  # mypy hint

        if self._fut is not None and not self._fut.done():
            self._lock.release()
            return

        while True:
            try:
                *_, self._cmd, self._qos, self._fut = self._que.get_nowait()
            except Empty:
                self._cmd = self._qos = self._fut = None
                self._lock.release()
                return

            self._cmd_tx_count = 0
            self._cmd_tx_limit = min(self._qos.max_retries, self.max_retry_limit) + 1

            assert isinstance(self._fut, asyncio.Future)  # mypy hint
            if self._fut.done():  # e.g. TimeoutError
                self._que.task_done()
                continue

            break

        self._lock.release()

        try:
            assert self._cmd is not None, f"{self}: Coding error"  # mypy hint
            self._send_cmd(self._cmd)
        finally:
            self._que.task_done()

    def _send_cmd(self, cmd: Command, is_retry: bool = False) -> None:
        """Wrapper to send a command with retries, until success or exception."""

        async def send_fnc_wrapper(cmd: Command) -> None:
            try:  # the wrapped function (actual Tx.write)
                await self._send_fnc(cmd)
            except exc.TransportError as err:
                self.set_state(IsInIdle, exception=err)

        # TODO: check what happens when exception here - why does it hang?
        assert cmd is not None, f"{self}: Coding error"

        try:  # the wrapped function (actual Tx.write)
            self._state.cmd_sent(cmd, is_retry=is_retry)
        except exc.ProtocolFsmError as err:
            self.set_state(IsInIdle, exception=err)
        else:
            self._loop.create_task(send_fnc_wrapper(cmd))


# With wait_for_reply=False
# AFTER. = <ProtocolContext state=IsInIdle>
# BEFORE = <ProtocolContext state=IsInIdle cmd_=2349|RQ|01:145038|08, tx_count=0/4>
# AFTER. = <ProtocolContext state=WantEcho cmd_=2349|RQ|01:145038|08, tx_count=1/4>
# BEFORE = <ProtocolContext state=WantEcho echo=2349|RQ|01:145038|08, tx_count=1/4>
# AFTER. = <ProtocolContext state=WantRply echo=2349|RQ|01:145038|08, tx_count=1/4>: wait_for_reply=False
#
# BEFORE = <ProtocolContext state=WantRply echo=2349|RQ|01:145038|08, tx_count=1/4>: result=2349|RQ|01:145038|08
# AFTER. = <ProtocolContext state=IsInIdle>

# With wait_for_reply=True
# AFTER. = <ProtocolContext state=IsInIdle>
# BEFORE = <ProtocolContext state=IsInIdle cmd_=0004|RQ|01:145038|05, tx_count=0/4>
# AFTER. = <ProtocolContext state=WantEcho cmd_=0004|RQ|01:145038|05, tx_count=1/4>
# BEFORE = <ProtocolContext state=WantEcho echo=0004|RQ|01:145038|05, tx_count=1/4>
# AFTER. = <ProtocolContext state=WantRply echo=0004|RQ|01:145038|05, tx_count=1/4>: wait_for_reply=True
#
# BEFORE = <ProtocolContext state=WantRply rply=0004|RP|01:145038|05, tx_count=1/4>: result=0004|RP|01:145038|05
# AFTER. = <ProtocolContext state=IsInIdle>

#######################################################################################

# NOTE: Because .dst / .src may switch from Address to Device from one pkt to the next:
#  - use: pkt.dst.id == self._echo_pkt.src.id
#  - not: pkt.dst    is self._echo_pkt.src


class ProtocolStateBase:
    def __init__(self, context: ProtocolContext) -> None:
        self._context = context

        self._sent_cmd: Command | None = None
        self._echo_pkt: Packet | None = None
        self._rply_pkt: Packet | None = None

    def __repr__(self) -> str:
        msg = f"<ProtocolState state={self.__class__.__name__}"
        if self._rply_pkt:
            return msg + f" rply={self._rply_pkt._hdr}>"
        if self._echo_pkt:
            return msg + f" echo={self._echo_pkt._hdr}>"
        if self._sent_cmd:
            return msg + f" cmd_={self._sent_cmd._hdr}>"
        return msg + ">"

    def connection_made(self) -> None:  # For all states except Inactive
        """Do nothing, as (except for InActive) we're already connected."""
        pass

    def connection_lost(self) -> None:  # Varies by states (not needed if Inactive)
        """Transition to Inactive, regardless of current state."""

        if isinstance(self._context._state, Inactive):
            return

        if isinstance(self._context._state, IsInIdle):
            self._context.set_state(Inactive)
            return

        self._context.set_state(
            Inactive, exception=exc.TransportError("Connection lost")
        )

    def pkt_rcvd(self, pkt: Packet) -> None:  # Different for each state
        """Raise a NotImplementedError."""
        raise NotImplementedError("Invalid state to receive a packet")

    def writing_paused(self) -> None:  # Currently same for all states (TBD)
        """Do nothing."""
        pass

    def writing_resumed(self) -> None:  # Currently same for all states (TBD)
        """Do nothing."""
        pass

    def cmd_sent(  # For all except IsInIdle, WantEcho
        self, cmd: Command, is_retry: bool | None = None
    ) -> None:
        raise exc.ProtocolFsmError(f"Invalid state to send a command: {self._context}")


class Inactive(ProtocolStateBase):
    """The Protocol is not connected to the transport layer."""

    def connection_made(self) -> None:
        """Transition to IsInIdle."""
        self._context.set_state(IsInIdle)

    def pkt_rcvd(self, pkt: Packet) -> None:  # raise ProtocolFsmError
        """Raise an exception, as a packet is not expected in this state."""

        assert self._sent_cmd is None, f"{self}: Coding error"

        if pkt.code != Code._PUZZ:
            _LOGGER.warning("%s: Invalid state to receive a packet", self._context)


class IsInIdle(ProtocolStateBase):
    """The Protocol is not in the process of sending a Command."""

    def pkt_rcvd(self, pkt: Packet) -> None:  # Do nothing
        """Do nothing as we're not expecting an echo, nor a reply."""

        assert self._sent_cmd is None, f"{self}: Coding error"

        pass

    def cmd_sent(  # Will expect an Echo
        self, cmd: Command, is_retry: bool | None = None
    ) -> None:
        """Transition to WantEcho."""

        assert self._sent_cmd is None and is_retry is False, f"{self}: Coding error"

        self._sent_cmd = cmd

        # HACK for headers with sentinel values:
        #  I --- 18:000730 18:222222 --:------ 30C9 003 000333  # 30C9| I|18:000730,    *but* will be: 30C9| I|18:222222
        #  I --- --:------ --:------ 18:000730 0008 002 00BB    # 0008| I|18:000730|00, *and* will be unchanged

        if HGI_DEVICE_ID in cmd.tx_header:  # HACK: what do I do about this
            cmd._hdr_ = cmd._hdr_.replace(HGI_DEVICE_ID, self._context._protocol.hgi_id)
        self._context.set_state(WantEcho)


class WantEcho(ProtocolStateBase):
    """The Protocol is waiting to receive an echo Packet."""

    # NOTE: unfortunately, the cmd's src / echo's src can be different:
    # RQ --- 18:000730 10:052644 --:------ 3220 005 0000050000  # RQ|10:048122|3220|05
    # RQ --- 18:198151 10:052644 --:------ 3220 005 0000050000  # RQ|10:048122|3220|05

    def __init__(self, context: ProtocolContext) -> None:
        super().__init__(context)

        self._sent_cmd = context._state._sent_cmd
        # if isinstance(context._state, WantEcho | WantRply):
        #     self._echo_pkt = context._state._echo_pkt
        # else:
        #     self._echo_pkt = None

    def pkt_rcvd(self, pkt: Packet) -> None:  # Check if pkt is expected Echo
        """If the pkt is the expected Echo, transition to IsInIdle, or WantRply."""

        # RQ --- 18:002563 01:078710 --:------ 2349 002 0200                # 2349|RQ|01:078710|02
        # RP --- 01:078710 18:002563 --:------ 2349 007 0201F400FFFFFF      # 2349|RP|01:078710|02
        #  W --- 30:257306 01:096339 --:------ 313F 009 0060002916050B07E7  # 313F| W|01:096339
        #  I --- 01:096339 30:257306 --:------ 313F 009 00FC0029D6050B07E7  # 313F| I|01:096339

        assert self._sent_cmd, f"{self}: Coding error"  # mypy hint

        # if self._sent_cmd.rx_header and pkt._hdr == self._sent_cmd.rx_header:
        #     _LOGGER.error("hdr=%s", self._sent_cmd.rx_header)
        #     _LOGGER.error("src=%s", self._sent_cmd.src.id)
        #     _LOGGER.error("dst=%s", pkt.dst.id)

        if (
            self._sent_cmd.rx_header
            and pkt._hdr == self._sent_cmd.rx_header
            and (
                pkt.dst.id == self._sent_cmd.src.id
                or (  # handle: 18:146440 == 18:000730
                    self._sent_cmd.src.id == HGI_DEVICE_ID
                    and pkt.dst.id == self._context._protocol.hgi_id
                )
            )
        ):
            _LOGGER.warning(
                "%s: Invalid state to receive a reply (expecting echo)", self._context
            )

            self._rply_pkt = pkt
            self._context.set_state(IsInIdle, result=pkt)
            return

        # HACK for packets with addr sets like (issue is only with sentinel values?):
        #  I --- --:------ --:------ 18:000730 0008 002 00BB

        if HGI_DEVICE_ID in pkt._hdr:  # HACK: what do I do about this?
            pkt__hdr = pkt._hdr_.replace(HGI_DEVICE_ID, self._context._protocol.hgi_id)
        else:
            pkt__hdr = pkt._hdr

        if pkt__hdr != self._sent_cmd.tx_header:
            return

        # # HACK: for testing - drop some packets
        # import random
        # if random.random() < 0.2:
        #     return

        self._echo_pkt = pkt
        if self._sent_cmd.rx_header:
            self._context.set_state(WantRply)
        else:
            self._context.set_state(IsInIdle, result=pkt)

    def cmd_sent(self, cmd: Command, is_retry: bool | None = None) -> None:
        """Transition to WantEcho (i.e. a retransmit)."""

        assert self._sent_cmd is not None and is_retry is True, f"{self}: Coding error"

        # NOTE: don't self._context.set_state(WantEcho) here - may cause endless loop


class WantRply(ProtocolStateBase):
    """The Protocol is waiting to receive an reply Packet."""

    # NOTE: is possible get a false rply (same rx_header), e.g.:
    # RP --- 10:048122 18:198151 --:------ 3220 005 00C0050000  # 3220|RP|10:048122|05
    # RP --- 10:048122 01:145038 --:------ 3220 005 00C0050000  # 3220|RP|10:048122|05

    # NOTE: unfortunately, the cmd's src / rply's dst can still be different:
    # RQ --- 18:000730 10:052644 --:------ 3220 005 0000050000  # 3220|RQ|10:048122|05
    # RP --- 10:048122 18:198151 --:------ 3220 005 00C0050000  # 3220|RP|10:048122|05

    def __init__(self, context: ProtocolContext) -> None:
        super().__init__(context)

        self._sent_cmd = context._state._sent_cmd
        self._echo_pkt = context._state._echo_pkt

    def pkt_rcvd(self, pkt: Packet) -> None:  # Check if pkt is expected Reply
        """If the pkt is the expected reply, transition to IsInIdle."""

        assert self._sent_cmd, f"{self}: Coding error"  # mypy hint
        assert self._echo_pkt, f"{self}: Coding error"  # mypy hint

        # NOTE: beware collisions: same header, but is not reply (must check RSSI or src)
        # 2024-04-16 08:28:33.895 000 RQ --- 18:146440 10:048122 --:------ 3220 005 0000110000  # 3220|RQ|10:048122|11
        # 2024-04-16 08:28:33.910 052 RQ --- 01:145038 10:048122 --:------ 3220 005 0000110000  # 3220|RQ|10:048122|11

        if pkt._hdr == self._sent_cmd.tx_header and pkt.src == self._echo_pkt.src:
            _LOGGER.warning(
                "%s: Invalid state to receive an echo (expecting reply)", self._context
            )
            return  # do not transition, wait until existing timer expires

        # HACK: special case: if null log entry for log_idx=nn, then
        # HACK: rx_hdr will be 0418|RP|01:145038|00, and not 0418|RP|01:145038|nn
        # HACK: wait_for_reply must be true for RQ|0418 commands
        if (
            self._sent_cmd.rx_header[:8] == "0418|RP|"  # type: ignore[index]
            and self._sent_cmd.rx_header[:-2] == pkt._hdr[:-2]  # type: ignore[index]
            and pkt.payload == "000000B0000000000000000000007FFFFF7000000000"
        ):
            self._rply_pkt = pkt

        elif pkt._hdr != self._sent_cmd.rx_header:
            return

        else:
            self._rply_pkt = pkt

        self._context.set_state(IsInIdle, result=pkt)


#######################################################################################


_ProtocolStateT: TypeAlias = Inactive | IsInIdle | WantEcho | WantRply

_ProtocolStateClassT: TypeAlias = (
    type[Inactive] | type[IsInIdle] | type[WantEcho] | type[WantRply]
)



# TODO: get_state()  # include_expired=False)
# TODO: get_state(include_expired=True)
# TODO: set_state(packets)  # , schema=None)
# TODO: set_state(packets, schema=test_schema)

# can use logs from test_systems



#!/usr/bin/env python3
"""RAMSES RF - Test the binding protocol with a virtual RF

NB: This test will likely fail with pytest -n x, because of the protocol's throttle
limits.
"""

import asyncio
from collections.abc import Callable
from typing import Any

import pytest
import serial  # type: ignore[import-untyped]

from ramses_rf import Message
from ramses_tx.const import SZ_ACTIVE_HGI, SZ_IS_EVOFW3, Code
from ramses_tx.protocol import RamsesProtocolT, create_stack, protocol_factory
from ramses_tx.schemas import DeviceIdT
from ramses_tx.transport import RamsesTransportT, transport_factory

from .virtual_rf import HgiFwTypes, VirtualRf

# ######################################################################################


async def assert_stack_state(
    protocol: RamsesProtocolT, transport: RamsesTransportT
) -> None:
    assert transport._this_pkt and transport._this_pkt.code == Code._PUZZ
    assert transport._this_pkt and transport._this_pkt.src.id == GWY_ID
    assert transport._prev_pkt is None

    assert transport.get_extra_info(SZ_ACTIVE_HGI) == GWY_ID
    assert transport.get_extra_info(SZ_IS_EVOFW3) is True

    assert protocol._active_hgi == GWY_ID
    assert protocol._is_evofw3 is True


def _msg_handler(msg: Message) -> None:
    pass


# ### TESTS ############################################################################


async def _test_create_stack(
    rf: VirtualRf,
    /,
    *,
    protocol_factory_: Callable | None = None,
    transport_factory_: Callable | None = None,
    disable_qos: bool | None = False,
    disable_sending: bool | None = False,
    enforce_include_list: bool = False,
    exclude_list: dict[DeviceIdT, dict] | None = None,
    include_list: dict[DeviceIdT, dict] | None = None,
    **kwargs: Any,  # TODO: these are for the transport_factory
) -> None:
    protocol: RamsesProtocolT
    transport: RamsesTransportT

    protocol, transport = await create_stack(
        _msg_handler,
        disable_qos=disable_qos,
        disable_sending=disable_sending,
        enforce_include_list=enforce_include_list,
        exclude_list=exclude_list,
        include_list=include_list,
        **kwargs,
    )

    try:
        await assert_stack_state(protocol, transport)
    except serial.SerialException as err:
        transport._close(exc=err)
        raise
    except (AssertionError, asyncio.InvalidStateError, TimeoutError):
        transport.close()
        raise
    else:
        transport.close()


async def _test_factories(
    rf: VirtualRf,
    /,
    *,
    protocol_factory_: Callable | None = None,
    transport_factory_: Callable | None = None,
    disable_qos: bool | None = False,
    disable_sending: bool | None = False,
    enforce_include_list: bool = False,
    exclude_list: dict[DeviceIdT, dict] | None = None,
    include_list: dict[DeviceIdT, dict] | None = None,
    **kwargs: Any,  # TODO: these are for the transport_factory
) -> None:
    protocol: RamsesProtocolT = protocol_factory(
        _msg_handler,
        disable_qos=disable_qos,
        disable_sending=disable_sending,
        enforce_include_list=enforce_include_list,
        exclude_list=exclude_list,
        include_list=include_list,
    )
    transport: RamsesTransportT = await transport_factory(
        protocol,
        disable_sending=disable_sending,
        **kwargs,
    )

    try:
        await assert_stack_state(protocol, transport)
    except serial.SerialException as err:
        transport._close(exc=err)
        raise
    except (AssertionError, asyncio.InvalidStateError, TimeoutError):
        transport.close()
        raise
    else:
        transport.close()


# ######################################################################################


GWY_ID = "18:111111"
OTH_ID = "18:123456"

TEST_SUITE_GOOD = {
    "00": {
        "include_list": {},
    },
    "01": {
        "enforce_include_list": True,
        "include_list": {GWY_ID: {"class": "HGI"}},
    },
    "02": {
        "enforce_include_list": True,
        "include_list": {GWY_ID: {"class": "HGI"}, OTH_ID: {"class": "HGI"}},
    },
    "03": {
        "enforce_include_list": True,
        "include_list": {OTH_ID: {"class": "HGI"}, GWY_ID: {"class": "HGI"}},
    },
    "04": {
        "enforce_include_list": True,
        "include_list": {OTH_ID: {"class": "HGI"}},
    },
}
TEST_SUITE_FAIL = {  # fails because exclude_list is checked before active_hgi
    "10": {
        "exclude_list": {GWY_ID: {"class": "HGI"}},
    },
    "11": {
        "enforce_include_list": True,
        "exclude_list": {GWY_ID: {"class": "HGI"}},
        "include_list": {GWY_ID: {"class": "HGI"}},
    },
}


@pytest.mark.xdist_group(name="virt_serial")
@pytest.mark.parametrize("idx", TEST_SUITE_GOOD)
async def test_create_stack(idx: str) -> None:
    """Check that Transport calls Protocol.connection_made() correctly."""

    rf = VirtualRf(2, start=True)
    rf.set_gateway(rf.ports[0], GWY_ID, fw_type=HgiFwTypes.EVOFW3)

    kwargs = {
        "port_name": rf.ports[0],
        "port_config": {},
        "extra": {"virtual_rf": rf},
    }

    try:
        await _test_create_stack(rf, **TEST_SUITE_GOOD[idx], **kwargs)
    finally:
        await rf.stop()


@pytest.mark.xdist_group(name="virt_serial")
@pytest.mark.parametrize("idx", TEST_SUITE_GOOD)
async def test_create_s_alt(idx: str) -> None:
    """Check that Transport calls Protocol.connection_made() correctly."""

    rf = VirtualRf(2, start=True)
    rf.set_gateway(rf.ports[0], GWY_ID, fw_type=HgiFwTypes.EVOFW3)

    kwargs = {
        "port_name": rf.ports[0],
        "port_config": {},
        "protocol_factory_": protocol_factory,
        "transport_factory_": transport_factory,
        "extra": {"virtual_rf": rf},
    }

    try:
        await _test_create_stack(rf, **TEST_SUITE_GOOD[idx], **kwargs)
    finally:
        await rf.stop()


@pytest.mark.xdist_group(name="virt_serial")
@pytest.mark.parametrize("idx", TEST_SUITE_GOOD)
async def test_factories_01(idx: str) -> None:
    """Check that Transport calls Protocol.connection_made() correctly.

    This is the method used by ramses_tf.gateway.py.
    """

    rf = VirtualRf(2, start=True)
    rf.set_gateway(rf.ports[0], GWY_ID, fw_type=HgiFwTypes.EVOFW3)

    kwargs = {
        "port_name": rf.ports[0],
        "port_config": {},
        "extra": {"virtual_rf": rf},
    }

    try:
        await _test_factories(rf, **TEST_SUITE_GOOD[idx], **kwargs)
    finally:
        await rf.stop()



#!/usr/bin/env python3
"""RAMSES RF - Expose an 0418 fault log (is a stateful process)."""

from __future__ import annotations

import dataclasses
import logging
from collections import OrderedDict
from typing import TYPE_CHECKING, NewType, TypeAlias

from ramses_tx import Command, Message, Packet
from ramses_tx.const import (
    SZ_LOG_ENTRY,
    SZ_LOG_IDX,
    FaultDeviceClass,
    FaultState,
    FaultType,
)
from ramses_tx.helpers import parse_fault_log_entry
from ramses_tx.schemas import DeviceIdT

from ramses_rf.const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    Code,
)

if TYPE_CHECKING:
    from ramses_rf.system.heat import _LogbookT


FaultTupleT: TypeAlias = tuple[FaultType, FaultDeviceClass, DeviceIdT | None, str]


DEFAULT_GET_LIMIT = 6


_LOGGER = logging.getLogger(__name__)


#  {'log_idx': '00', 'log_entry': ('21-12-23T11:59:35', 'restore', 'battery_low', 'actuator', '00', '04:164787', 'B0', '0000', 'FFFF7000')}


@dataclasses.dataclass(frozen=True, kw_only=True, order=True)
class FaultLogEntry:
    """A fault log entry of an evohome fault log.

    Fault log entries do have a log_idx attr, but this is merely their current location
    in the system's fault log.
    """

    timestamp: str  # #               # 21-12-23T11:59:35 - assume is unique
    fault_state: FaultState  # #      # fault, restore, unknown_c0
    fault_type: FaultType  # #        # system_fault, battery_low, sensor_fault, etc.
    domain_idx: str  # #              # 00-0F, FC, etc. ? only if dev_class is/not CTL?
    device_class: FaultDeviceClass  # # controller, actuator, sensor, etc.
    device_id: DeviceIdT | None  # #  # 04:164787

    # def __post_init__(self):
    #     def modify(device_id: DeviceIdT) -> DeviceIdT:
    #     object.__setattr__(self, "device_id", modify(self.device_id))

    def __str__(self) -> str:
        return (
            f"{self.timestamp}, {(self.fault_state + ','):<8} {self.fault_type}, "
            f"{self.device_id}, {self.domain_idx}, {self.device_class}"
        )

    def _is_matching_pair(self, other: object) -> bool:
        """Return True if the other entry could be a matching pair (fault/restore)."""

        if not isinstance(other, FaultLogEntry):  # TODO: make a parochial exception
            raise TypeError(f"{other} is not not a FaultLogEntry")

        if self.fault_state == FaultState.FAULT:
            return (
                other.fault_state == FaultState.RESTORE
                and self._as_tuple() == other._as_tuple()
                and other.timestamp > self.timestamp
            )

        if self.fault_state == FaultState.RESTORE:
            return (
                other.fault_state == FaultState.FAULT
                and self._as_tuple() == other._as_tuple()
                and other.timestamp < self.timestamp
            )

        return False

    def _as_tuple(self) -> FaultTupleT:  # only for use within this class
        """Return the log entry as a tuple, excluding dtm & state (fault/restore)."""

        return (
            self.fault_type,
            self.device_class,
            self.device_id,
            self.domain_idx,
        )

    @classmethod
    def from_msg(cls, msg: Message) -> FaultLogEntry:
        """Create a fault log entry from a message's packet."""
        return cls.from_pkt(msg._pkt)

    @classmethod
    def from_pkt(cls, pkt: Packet) -> FaultLogEntry:
        """Create a fault log entry from a packet's payload."""

        log_entry = parse_fault_log_entry(pkt.payload)
        if log_entry is None:  # TODO: make a parochial exception
            raise TypeError("Null fault log entry")

        return cls(**{k: v for k, v in log_entry.items() if k[:1] != "_"})  # type: ignore[arg-type]


FaultDtmT = NewType("FaultDtmT", str)
FaultIdxT = NewType("FaultIdxT", int)

FaultLogT: TypeAlias = dict[FaultDtmT, FaultLogEntry]
FaultMapT: TypeAlias = OrderedDict[FaultIdxT, FaultDtmT]


class FaultLog:  # 0418  # TODO: use a NamedTuple
    """The fault log of an evohome system.

    This code assumes that the `timestamp` attr of each log entry is a unique identifer.

    Null entries do not have a timestamp. All subsequent entries will also be null.

    The `log_idx` is not an identifier: it is merely the current position of a log entry
    in the system log.

    New entries are added to the top of the log (log_idx=0), and the log_idx is
    incremented for all exisiting log enties.
    """

    _MAX_LOG_IDX = 0x3E

    def __init__(self, tcs: _LogbookT) -> None:
        self._tcs: _LogbookT = tcs
        self.id = tcs.id
        self._gwy = tcs._gwy

        self._log: FaultLogT = dict()
        self._map: FaultMapT = OrderedDict()
        self._log_done: bool | None = None

        self._is_current: bool = False  # if we now our log is out of date
        self._is_getting: bool = False

    def _insert_into_map(self, idx: FaultIdxT, dtm: FaultDtmT | None) -> FaultMapT:
        """Rebuild the map (as best as possible), given the a log entry."""

        new_map: FaultMapT = OrderedDict()

        # usu. idx == 0, but could be > 0
        new_map |= {
            k: v for k, v in self._map.items() if k < idx and (dtm is None or v > dtm)
        }

        if dtm is None:  # there are no subsequent log entries
            return new_map

        new_map |= {idx: dtm}

        if not (idxs := [k for k, v in self._map.items() if v < dtm]):
            return new_map

        if (next_idx := min(idxs)) > idx:
            diff = 0
        elif next_idx == idx:
            diff = 1  # next - idx + 1
        else:
            diff = idx + 1  # 1 if self._map.get(idx) else 0

        new_map |= {
            k + diff: v  # type: ignore[misc]
            for k, v in self._map.items()
            if (k >= idx or v < dtm) and k + diff <= self._MAX_LOG_IDX
        }

        return new_map

    def handle_msg(self, msg: Message) -> None:
        """Handle a fault log message (some valid payloads should be ignored)."""

        assert msg.code == Code._0418 and msg.verb in (I_, RP), "Coding error"

        if msg.verb == RP and msg.payload[SZ_LOG_ENTRY] is None:
            # such payloads have idx == "00" (is sentinel for null), so can't know the
            # correspondings RQ's log_idx, but if verb == I_, safely assume log_idx is 0
            return

        self._process_msg(msg)

    def _process_msg(self, msg: Message) -> None:
        """Handle a processable fault log message."""

        if msg.verb == I_:
            self._is_current = False

        if SZ_LOG_IDX not in msg.payload:
            return  # we can't do anything useful with this message

        idx: FaultIdxT = int(msg.payload[SZ_LOG_IDX], 16)  # type: ignore[assignment]

        if msg.payload[SZ_LOG_ENTRY] is None:  # NOTE: Subsequent entries will be empty
            self._map = self._insert_into_map(idx, None)
            self._log = {k: v for k, v in self._log.items() if k in self._map.values()}
            return  # If idx != 0, should we also check from idx = 0?

        entry = FaultLogEntry.from_msg(msg)  # if msg.payload[SZ_LOG_ENTRY] else None
        dtm: FaultDtmT = entry.timestamp  # type: ignore[assignment]

        if self._map.get(idx) == dtm:
            return  # i.e. No evidence anything has changed

        if dtm not in self._log:
            self._log |= {dtm: entry}  # must add entry before _insert_into_map()
        self._map = self._insert_into_map(idx, dtm)  # updates self._map
        self._log = {k: v for k, v in self._log.items() if k in self._map.values()}

        # if idx != 0:  # there's other (new/changed) entries above this one?
        #     pass

    def _hack_pkt_idx(self, pkt: Packet, cmd: Command) -> Message:
        """Modify the Packet so that it has the log index of its corresponding Command.

        If there is no log entry for log_idx=<idx>, then the headers wont match:
        - cmd rx_hdr is 0418|RP|<ctl_id>|<idx> (expected)
        - pkt hdr will  0418|RP|<ctl_id>|00    (response from controller)

        We can only assume that the Pkt is the reply to the Cmd, which is why using
        QoS with wait_for_reply=True is vital when getting fault log entries.

        We can assume 0418| I|<ctl_id>|00 is only for log_idx=00 (I|0418s are stateless)
        """

        assert pkt.verb == RP and pkt.code == Code._0418 and pkt._idx == "00"
        assert pkt.payload == "000000B0000000000000000000007FFFFF7000000000"

        assert cmd.verb == RQ and pkt.code == Code._0418
        assert cmd.rx_header and cmd.rx_header[:-2] == pkt._hdr[:-2]  # reply to this RQ

        if cmd._idx == "00":  # no need to hack
            return Message(pkt)

        idx = cmd.rx_header[-2:]  # cmd._idx could be bool/None?
        pkt.payload = f"0000{idx}B0000000000000000000007FFFFF7000000000"

        # NOTE: must now reset pkt payload, and its header
        pkt._repr = pkt._hdr_ = pkt._ctx_ = pkt._idx_ = None  # type: ignore[assignment]
        pkt._frame = pkt._frame[:50] + idx + pkt._frame[52:]

        assert pkt._hdr == cmd.rx_header, f"{self}: Coding error"
        assert (
            str(pkt) == pkt._frame[:50] + idx + pkt._frame[52:]
        ), f"{self}: Coding error"

        msg = Message(pkt)
        msg._payload = {SZ_LOG_IDX: idx, SZ_LOG_ENTRY: None}  # PayDictT._0418_NULL

        return msg

    async def get_faultlog(
        self,
        /,
        *,
        start: int = 0,
        limit: int | None = DEFAULT_GET_LIMIT,
        force_refresh: bool = False,
    ) -> dict[FaultIdxT, FaultLogEntry]:
        """Retrieve the fault log from the controller."""

        if limit is None:
            limit = DEFAULT_GET_LIMIT

        self._is_getting = True

        for idx in range(start, min(start + limit, 64)):
            cmd = Command.get_system_log_entry(self.id, idx)
            pkt = await self._gwy.async_send_cmd(cmd, wait_for_reply=True)

            if pkt.payload == "000000B0000000000000000000007FFFFF7000000000":
                msg = self._hack_pkt_idx(pkt, cmd)  # RPs for null entrys have idx=="00"
                self._process_msg(msg)  # since pkt via dispatcher aint got idx
                break
            self._process_msg(Message(pkt))  # JIC dispatcher doesn't do this for us

        self._is_current = False
        self._is_getting = False

        return self.faultlog

    @property
    def faultlog(self) -> dict[FaultIdxT, FaultLogEntry]:
        """Return the fault log of a system."""

        # if self._faultlog:
        #     return self._faultlog

        return {idx: self._log[dtm] for idx, dtm in self._map.items()}

    def is_current(self, force_io: bool | None = None) -> bool:
        """Return True if the local fault log is identical to the controllers.

        If force_io, retrieve the 0th log entry and check it is identical to the local
        copy.
        """

        if not self._is_current:
            return False
        return True

    @property
    def latest_event(self) -> FaultLogEntry | None:
        """Return the most recently logged event (fault or restore), if any."""

        if not self._log:  # TODO: raise exception or retrive log (make function)?
            return None

        return self._log[max(k for k in self._log)]

    @property
    def latest_fault(self) -> FaultLogEntry | None:
        """Return the most recently logged fault, if any."""

        if not self._log:  # TODO: raise exception or retrive log (make function)?
            return None

        faults = [k for k, v in self._log.items() if v.fault_state == FaultState.FAULT]

        if not faults:
            return None

        return self._log[max(faults)]

    @property
    def active_faults(self) -> tuple[FaultLogEntry, ...] | None:
        """Return a list of all faults outstanding (i.e. no corresponding restore)."""

        if not self._log:  # TODO: raise exception or retrive log (make function)?
            return None

        restores = {}
        faults = {}

        for entry in sorted(self._log.values(), reverse=True):
            if entry.fault_state == FaultState.RESTORE:
                # keep to match against upcoming faults
                restores[entry._as_tuple()] = entry

            if entry.fault_state == FaultState.FAULT:
                # look for (existing) matching restore, otherwise keep
                if entry._as_tuple() in restores:
                    del restores[entry._as_tuple()]
                else:
                    faults[entry._as_tuple()] = entry

        return tuple(faults.values())



#!/usr/bin/env python3
"""RAMSES RF - Test the Faultlog functions."""

import random
from datetime import datetime as dt
from typing import Any

from ramses_rf import Address, Command, Message, Packet
from ramses_rf.system.faultlog import FaultLog, FaultLogEntry
from ramses_tx.address import HGI_DEVICE_ID
from ramses_tx.const import SZ_LOG_ENTRY, FaultDeviceClass, FaultState, FaultType
from ramses_tx.schemas import DeviceIdT
from ramses_tx.typed_dicts import LogIdxT

from ramses_tx.const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    Code,
)

from .helpers import TEST_DIR

WORK_DIR = f"{TEST_DIR}/parsers"


# ### TEST DATA #######################################################################

CTL_ID = Address("01:145038").id
HGI_ID = HGI_DEVICE_ID


def _fault_log_entry(
    *args: Any, timestamp: str | None = None, **kwargs: Any
) -> FaultLogEntry:
    if timestamp is None:
        timestamp = dt.strftime(dt.now(), "%y-%m-%dT%H:%M:%S")

    return FaultLogEntry(
        timestamp=timestamp,
        fault_state=args[0],
        fault_type=args[1],
        domain_idx=kwargs.get("domain_idx", "00"),
        device_class=args[2],
        device_id=kwargs.get("device_id"),
    )


# Keys corresponding to order in the faultlog
TEST_FAULTS: dict[LogIdxT, FaultLogEntry] = {}

TEST_FAULTS["00"] = _fault_log_entry(
    FaultState.RESTORE,
    FaultType.BATTERY_ERROR,
    FaultDeviceClass.CONTROLLER,
    timestamp="21-12-23T00:59:00",
)
TEST_FAULTS["01"] = _fault_log_entry(
    FaultState.FAULT,
    FaultType.BATTERY_ERROR,
    FaultDeviceClass.CONTROLLER,
    timestamp="21-12-23T00:58:01",
)
TEST_FAULTS["02"] = _fault_log_entry(
    FaultState.RESTORE,
    FaultType.BATTERY_LOW,
    FaultDeviceClass.ACTUATOR,
    timestamp="21-12-23T00:57:02",
    device_id="04:111111",
    domain_idx="03",
)
TEST_FAULTS["03"] = _fault_log_entry(
    FaultState.FAULT,
    FaultType.BATTERY_LOW,
    FaultDeviceClass.ACTUATOR,
    timestamp="21-12-23T00:56:03",
    device_id="04:111111",
    domain_idx="03",
)
TEST_FAULTS["04"] = _fault_log_entry(
    FaultState.RESTORE,
    FaultType.COMMS_FAULT,
    FaultDeviceClass.SETPOINT,
    timestamp="21-12-23T00:55:04",
    device_id="03:123456",
    domain_idx="06",
)
TEST_FAULTS["05"] = _fault_log_entry(
    FaultState.FAULT,
    FaultType.COMMS_FAULT,
    FaultDeviceClass.SETPOINT,
    timestamp="21-12-23T00:54:05",
    device_id="03:123456",
    domain_idx="06",
)

EXPECTED_MAP = {
    0: "21-12-23T00:59:00",
    1: "21-12-23T00:58:01",
    2: "21-12-23T00:57:02",
    3: "21-12-23T00:56:03",
    4: "21-12-23T00:55:04",
    5: "21-12-23T00:54:05",
}


# ### FIXTURES, ETC ###################################################################


class EvohomeStub:
    def __init__(self, ctl_id: DeviceIdT) -> None:
        self.id = ctl_id
        self._gwy = None


def _proc_log_line(log_line: str) -> None:
    try:
        pkt = Packet.from_file(log_line[:26], log_line[27:])
    except ValueError:
        return

    msg = Message(pkt)
    if msg.code != Code._0418 or not msg.payload.get(SZ_LOG_ENTRY):
        return

    entry = FaultLogEntry.from_msg(msg)
    assert entry


def _proc_null_fault_entry(fault_log: FaultLog, _log_idx: LogIdxT = "00") -> None:
    """Return a 0418 packet with no entry."""
    cmd = Command.from_attrs(
        I_, CTL_ID, Code._0418, f"0000{_log_idx}B0000000000000000000007FFFFF7000000000"
    )
    fault_log.handle_msg(Message(Packet._from_cmd(cmd)))


def _proc_test_fault_entry(
    fault_log: FaultLog, text_idx: LogIdxT, _log_idx: LogIdxT = "00"
) -> None:
    entry: FaultLogEntry = TEST_FAULTS[text_idx]

    cmd = Command._put_system_log_entry(
        CTL_ID,
        entry.fault_state,
        entry.fault_type,
        entry.device_class,
        device_id=entry.device_id,
        domain_idx=entry.domain_idx,
        _log_idx=_log_idx,
        timestamp=entry.timestamp,
    )
    fault_log.handle_msg(Message(Packet._from_cmd(cmd)))


# ### TESTS ###########################################################################


def test_faultlog_entries() -> None:
    """Test instantiation of faultlog entries."""

    with open(f"{WORK_DIR}/code_0418.log") as f:
        while line := (f.readline()):
            _proc_log_line(line)


def test_faultlog_instantiation_0() -> None:
    """Log entries arrive in order of timestamp (i.e. as they'd occur)."""

    fault_log = FaultLog(EvohomeStub(CTL_ID))  # type: ignore[type-var]

    # log entries arrive in order of timestamp (i.e. as they'd occur)
    for i in reversed(range(len(TEST_FAULTS))):
        _proc_test_fault_entry(fault_log, f"{i:02}")  # _log_idx="00")

    # assert sorted(fault_log._log.keys(), reverse=True) == list(EXPECTED_MAP.values())
    assert fault_log._map == EXPECTED_MAP


def test_faultlog_instantiation_1() -> None:
    """Log entries arrive in order of log_idx (e.g. enumerating the log via RQs)."""

    fault_log = FaultLog(EvohomeStub(CTL_ID))  # type: ignore[type-var]

    # log entries arrive in order of log_idx (e.g. enumerating the log via RQs)
    for i in reversed(range(len(TEST_FAULTS))):
        _proc_test_fault_entry(fault_log, f"{i:02}", _log_idx=f"{i:02}")

    assert sorted(fault_log._log.keys(), reverse=True) == list(EXPECTED_MAP.values())
    assert fault_log._map == EXPECTED_MAP


def test_faultlog_instantiation_2() -> None:
    """Log entries arrive in random order albeit with their correct log_idx."""

    fault_log = FaultLog(EvohomeStub(CTL_ID))  # type: ignore[type-var]

    # log entries arrive in random order albeit with their correct log_idx
    numbers = list(range(len(TEST_FAULTS)))
    random.shuffle(numbers)

    for i in numbers:
        _proc_test_fault_entry(fault_log, f"{i:02}", _log_idx=f"{i:02}")

    assert sorted(fault_log._log.keys(), reverse=True) == list(EXPECTED_MAP.values())
    assert fault_log._map == EXPECTED_MAP


def test_faultlog_instantiation_3() -> None:
    """Log entries arrive in an order set to confuse."""

    fault_log = FaultLog(EvohomeStub(CTL_ID))  # type: ignore[type-var]

    # a log with two entries arrives in order
    _proc_test_fault_entry(fault_log, "05")
    _proc_test_fault_entry(fault_log, "04")

    assert fault_log._map == {
        0: "21-12-23T00:55:04",
        1: "21-12-23T00:54:05",
    }

    # the two entries arrives out of order
    _proc_test_fault_entry(fault_log, "05", _log_idx="01")
    _proc_test_fault_entry(fault_log, "04", _log_idx="00")

    assert fault_log._map == {
        0: "21-12-23T00:55:04",
        1: "21-12-23T00:54:05",
    }

    # the log is cleared
    _proc_null_fault_entry(fault_log)

    assert fault_log._map == {}

    # a log with three entries is enumerated, kinda
    _proc_test_fault_entry(fault_log, "03", _log_idx="00")
    # roc_fault_entry(fault_log, "04", _log_idx="01")  # went missing
    _proc_test_fault_entry(fault_log, "05", _log_idx="02")

    assert fault_log._map == {
        0: "21-12-23T00:56:03",
        2: "21-12-23T00:54:05",
    }

    # the missing entry arrives, only after a new entry
    _proc_test_fault_entry(fault_log, "02")  # pushes others down
    _proc_test_fault_entry(fault_log, "01")  # pushes others down
    _proc_test_fault_entry(fault_log, "04", _log_idx="03")  # _log_idx was 01, above

    assert fault_log._map == {
        0: "21-12-23T00:58:01",
        1: "21-12-23T00:57:02",
        2: "21-12-23T00:56:03",
        3: "21-12-23T00:55:04",
        4: "21-12-23T00:54:05",
    }

    # a new entry
    _proc_test_fault_entry(fault_log, "00")  # pushes others down

    assert fault_log._map == {
        0: "21-12-23T00:59:00",
        1: "21-12-23T00:58:01",
        2: "21-12-23T00:57:02",
        3: "21-12-23T00:56:03",
        4: "21-12-23T00:55:04",
        5: "21-12-23T00:54:05",
    }


def test_faultlog_instantiation_4() -> None:
    """Log entries arrive in an order set to confuse."""

    fault_log = FaultLog(EvohomeStub(CTL_ID))  # type: ignore[type-var]

    # a log with three entries is enumerated, kinda
    _proc_test_fault_entry(fault_log, "03", _log_idx="00")
    # roc_fault_entry(fault_log, "04", _log_idx="01")  # went missing
    _proc_test_fault_entry(fault_log, "05", _log_idx="02")

    assert fault_log._map == {
        0: "21-12-23T00:56:03",
        2: "21-12-23T00:54:05",
    }

    _proc_test_fault_entry(fault_log, "02", _log_idx="02")  # pushes others down

    assert fault_log._map == {
        2: "21-12-23T00:57:02",
        3: "21-12-23T00:56:03",
        5: "21-12-23T00:54:05",
    }

    _proc_test_fault_entry(fault_log, "01")

    assert fault_log._map == {
        0: "21-12-23T00:58:01",
        2: "21-12-23T00:57:02",
        3: "21-12-23T00:56:03",
        5: "21-12-23T00:54:05",
    }

    _proc_test_fault_entry(fault_log, "01", _log_idx="01")

    assert fault_log._map == {
        1: "21-12-23T00:58:01",
        2: "21-12-23T00:57:02",
        3: "21-12-23T00:56:03",
        5: "21-12-23T00:54:05",
    }

    _proc_test_fault_entry(fault_log, "04", _log_idx="04")  # _log_idx was 01, above

    assert fault_log._map == {
        1: "21-12-23T00:58:01",
        2: "21-12-23T00:57:02",
        3: "21-12-23T00:56:03",
        4: "21-12-23T00:55:04",
        5: "21-12-23T00:54:05",
    }



#!/usr/bin/env python3
"""A virtual RF network useful for testing."""

from enum import StrEnum
from typing import TypedDict


class _ComPortsT(TypedDict):
    manufacturer: str
    product: str
    vid: int
    pid: int

    description: str
    serial_number: str | None
    interface: str | None

    device: str
    name: str


# NOTE: Below values are from real devices (with some contrived values)

COMPORTS_ATMEGA32U4: _ComPortsT = {  # 8/16 MHz atmega32u4 (HW Uart)
    "manufacturer": "SparkFun",
    "product": "evofw3 atmega32u4",
    "vid": 0x1B4F,  # aka SparkFun Electronics
    "pid": 0x9206,
    #
    "description": "evofw3 atmega32u4",
    "serial_number": None,
    "interface": None,
    #
    "device": "/dev/ttyACM0",  # is not a fixed value
    "name": "ttyACM0",  # not fixed
}

COMPORTS_ATMEGA328P: _ComPortsT = {  # 16MHZ atmega328 (SW Uart)
    "manufacturer": "FTDI",
    "product": "FT232R USB UART",
    "vid": 0x0403,  # aka Future Technology Devices International Ltd.
    "pid": 0x6001,
    #
    "description": "FT232R USB UART - FT232R USB UART",
    "serial_number": "A50285BI",
    "interface": "FT232R USB UART",
    #
    "device": "/dev/ttyUSB0",  # is not a fixed value
    "name": "ttyUSB0",  # not fixed
}

COMPORTS_TI4310: _ComPortsT = {  # Honeywell HGI80 (partially contrived)
    "manufacturer": "Texas Instruments",
    "product": "TUSB3410 Boot Device",
    "vid": 0x10AC,  # aka Honeywell, Inc.
    "pid": 0x0102,
    #
    "description": "TUSB3410 Boot Device",  # contrived
    "serial_number": "TUSB3410",
    "interface": None,  # assumed
    #
    "device": "/dev/ttyUSB0",  # is not a fixed value
    "name": "ttyUSB0",  # not fixed
}


class HgiFwTypes(StrEnum):
    EVOFW3 = " ".join(COMPORTS_ATMEGA32U4[k] for k in ("manufacturer", "product"))  # type: ignore[literal-required]
    HGI_80 = " ".join(COMPORTS_TI4310[k] for k in ("manufacturer", "product"))  # type: ignore[literal-required]


SCHEMA_1 = {
    "orphans_hvac": ["41:111111"],
    "known_list": {
        "18:111111": {"class": "HGI", "fw_version": "EVOFW3"},
        "41:111111": {"class": "REM"},
    },
}

SCHEMA_2 = {
    "orphans_hvac": ["42:222222"],
    "known_list": {
        "18:222222": {"class": "HGI", "fw_version": "HGI_80"},
        "42:222222": {"class": "FAN"},
    },
}

SCHEMA_3 = {
    "orphans_hvac": ["42:333333"],
    "known_list": {"18:333333": {"class": "HGI"}, "42:333333": {"class": "FAN"}},
}
